Reforgerade = LibStub("AceAddon-3.0"):NewAddon("Reforgerade", "AceConsole-3.0", "AceTimer-3.0")
local AceGUI = LibStub("AceGUI-3.0")
local Reforgerade = Reforgerade
local reforgeString
local reforgeTable = {}
local RfInfo = LibStub("LibReforgingInfo-1.0")

local debugPrint = false
local debugErrors = true

Reforgerade:RegisterChatCommand("rfa", function() ReforgeradeInputFrame.frame:Show() end)
Reforgerade:RegisterChatCommand("reforgerade", function() ReforgeradeInputFrame.frame:Show() end)

--Stole from LibReforge
local SPI = 1
local DODGE = 2
local PARRY = 3
local HIT = 4
local CRIT = 5
local HASTE = 6
local EXP = 7
local MASTERY = 8

--Stole from LibReforge
local StatNames = {
	ITEM_MOD_SPIRIT_SHORT,
	ITEM_MOD_DODGE_RATING_SHORT,
	ITEM_MOD_PARRY_RATING_SHORT,
	ITEM_MOD_HIT_RATING_SHORT,
	ITEM_MOD_CRIT_RATING_SHORT,
	ITEM_MOD_HASTE_RATING_SHORT,
	ITEM_MOD_EXPERTISE_RATING_SHORT,
	ITEM_MOD_MASTERY_RATING_SHORT
}
StatNames[0] = "None"

local reforgequeue = {}
local reforgeParsers = {}
local waitingToUnforge = {}
local waitingToReforge = {}

local specialItems = {}
specialItems[select(1,GetItemInfo(63488)) or "Mistral Circle"] = true -- Let WoW localize the item names for us
specialItems[select(1,GetItemInfo(63493)) or "Wind Stalker Leggings"] = true
specialItems[select(1,GetItemInfo(63503)) or "Wind Stalker Belt"] = true
specialItems[select(1,GetItemInfo(68127)) or "Stormwake the Tempest's Reach"] = true

function Reforgerade:DebugPrint(debugString)
	if (debugPrint) then
		self:Print("DEBUG: " .. debugString)
	end
end

function Reforgerade:ErrorPrint(errorString)
	if (debugErrors) then
		self:Print("|cffff0000 ERROR: " .. errorString .. "|r")
	end
end


function Reforgerade:BuildReforgeTable()
	local reforgeString = ReforgeradeInputFrame.editBox:GetText()
	reforgeTable = {}
	local parseAttempts = {}
	local best = 0
	local parserName = nil
	local pickedParser = nil
	if next(reforgeParsers) ~= nil then
		for parserName, parserFunction in pairs(reforgeParsers) do --for each parser added to our list let them try parsing the data
			parseAttempt = parserFunction(reforgeString)
			local validItems = 0
			if (parseAttempt and next(parseAttempt) ~= nil) then
				for itemName,reforgeID in pairs(parseAttempt) do
					if (GetItemInfo(itemName) or specialItems[itemName]) then
						if (reforgeID[1] and reforgeID[1]<= 8 and reforgeID[2] and reforgeID[2]<= 8) then
							if (reforgeID[1] and reforgeID[2]) then
--~ 								self:DebugPrint(parserName .. " says reforge " .. itemName .. " from " ..  StatNames[reforgeID[1]]  .. " to " .. StatNames[reforgeID[2]])
								validItems = validItems + 1
							end
						else
							self:ErrorPrint("Invalid reforge IDs for " .. itemName .. " returned by parser " .. parserName)
						end
					else
						self:ErrorPrint(itemName .. " appears to be an invalid item.")
					end
				end
			end
			self:DebugPrint(parserName .. " returned " .. validItems .. " valid item reforges")
			if (validItems > best) then
				best = validItems
				reforgeTable = parseAttempt
				pickedParser = parserName
			end
		end
		if next(reforgeTable) ~= nil then
			self:DebugPrint("Picked parser: " .. pickedParser)
		else
			self:ErrorPrint("Unable to parse input")
		end
--~ 		for k,v in pairs(reforgeTable) do
--~ 			print("Finalized Reforge " .. k .. " from " .. v[1] .. " to " .. v[2])
--~ 		end
	else
		self:ErrorPrint("No reforge parsers found!")
	end
end

function Reforgerade:ReforgeScan(silent)
	silent = silent or false --default to false if called with nil
--~ 	if ReforgeradeInputFrame.frame:IsShown() == false then ReforgeradeInputFrame.frame:Show() return end
	Reforgerade:BuildReforgeTable() --rebuild the list of want reforges
	reforgequeue = {} --clear the reforge queue
	local itemChanged = false
	for i = 1, 18 do
--~ 	for i = 1, 2 do
		local item = GetInventoryItemLink("player", i)
		if (item and i~=4) then --if we have and item equiped in the slot and the slot is not the shirt or tabard slot
			local itemName =Reforgerade:FixItemName(select(1,GetItemInfo(item)))
	--~ 		self:DebugPrint(itemName .. "|")
			local WantReforgeFrom, WantReforgeTo
			if (reforgeTable[itemName] ~= nil) then
				WantReforgeFrom = reforgeTable[itemName][1]
				WantReforgeTo = reforgeTable[itemName][2]
				local CurrentReforgeFrom, CurrentReforgeTo
				if RfInfo:IsItemReforged(item) then
					CurrentReforgeFrom, CurrentReforgeTo = RfInfo:GetReforgedStatIDs(RfInfo:GetReforgeID(item))			
				end
				if (CurrentReforgeFrom == nil or CurrentReforgeTo == nil) then
					CurrentReforgeFrom = 0
					CurrentReforgeTo = 0
				end
		--~ 			self:Print(itemName .. " is reforged from " .. CurrentReforgeFrom .. " to " .. CurrentReforgeTo)
				if (CurrentReforgeFrom ~= WantReforgeFrom or CurrentReforgeTo ~= WantReforgeTo) then
					if (not silent) then
						self:Print(item .. " currently reforged from " .. StatNames[CurrentReforgeFrom] .. " -> " .. StatNames[CurrentReforgeTo] .. " needs to be reforged from " .. StatNames[WantReforgeFrom] .. " -> " .. StatNames[WantReforgeTo])
						itemChanged = true
					end
					tinsert(reforgequeue, {i, WantReforgeFrom, WantReforgeTo})
				else
					if (not silent) then
--~ 	 					self:Print("No need to change reforging for " ..item)
					end
				end
			else
				if (not silent) then
--~ 	 				self:Print("Unable to find wanted reforge information for equiped item " .. item)
				end
			end
		end
	end
	if (not silent) then
		if (itemChanged == false) then
			self:Print("Everything is already reforged.")
		end
	end
end

local currentSlot = 0

function Reforgerade:ReforgeIt()
	Reforgerade:ReforgeScan(true)
--~ 	self:DebugPrint("Running ReforgingIt")
	if (ReforgingFrame == nil  or ReforgingFrame:IsVisible() == false) then
		self:ErrorPrint("Open the reforging frame")
		return
	end
	local clearReforge = false
	for k, v in pairs(reforgequeue) do
--~ 		self:Print(v[1])
		--Pick up the item and place it in the reforge windo
		local itemSlot = v[1]
		local item = GetInventoryItemLink("player", itemSlot)
		if (currentSlot ~= itemSlot) then
			ClearCursor()
			ReforgingFrameItemButton:Click() -- pick up the old item
			ClearCursor() --get rid of it
			PickupInventoryItem(itemSlot)
			ReforgingFrameItemButton:Click()
			currentSlot = itemSlot
		end
		
		--Remove the old reforge
		
		if (ReforgingFrameRestoreButton:IsEnabled()) then
			if (waitingToUnforge[itemSlot] == nil or waitingToUnforge[itemSlot] == false) then -- only try to unforge an item if we are not already waitng for it to unforge
				self:DebugPrint("Unforging " .. item)
--~ 				waitingToUnforge[itemSlot] = true
				ReforgingFrameRestoreButton:Click()
				Reforgerade:RetryReforge()
				return
			else
				self:DebugPrint("Still waiting for " .. item .. " to Unforge")
				Reforgerade:RetryReforge()
				return
			end
			clearReforge = true
		else
			waitingToUnforge[itemSlot] = false
			if (waitingToReforge[itemSlot] == nil or waitingToReforge[itemSlot] == false) then
				--Select First Stat
				ReforgingFrameFilterOldStatButton:Click()
				local reforgeStatButtons = {
					DropDownList1Button1,
					DropDownList1Button2,
				}
				for i = 1, 2 do
					ReforgeToOption = select(3, string.find(reforgeStatButtons[i]:GetText(), "%-?%+?%d*%s*(.*)"))
	--~ 			 	self:Print(ReforgeToOption)
					if (ReforgeToOption == StatNames[v[2]]) then
						reforgeStatButtons[i]:Click()
						break
					end
				end
				
				--Select Second Stat
				ReforgingFrameFilterNewStatButton:Click()
				local reforgeStatButtons = {
					DropDownList1Button1,
					DropDownList1Button2,
					DropDownList1Button3,
					DropDownList1Button4,
					DropDownList1Button5,
					DropDownList1Button6,
					DropDownList1Button7,
				}
				for i = 1, 7 do
					ReforgeToOption = select(3, string.find(reforgeStatButtons[i]:GetText(), "%-?%+?%d*%s*(.*)"))
	--~ 			 	self:Print(ReforgeToOption)
					if (ReforgeToOption == StatNames[v[3]]) then
						reforgeStatButtons[i]:Click()
--~ 						waitingToReforge[itemSlot] = true
						self:DebugPrint("Reforging " .. item)
						ReforgingFrameReforgeButton:Click() --accept the reforge
						Reforgerade:RetryReforge()
						return
					end
				end
			else
				self:DebugPrint("Still waiting for " .. item .. " to reforge")
				Reforgerade:RetryReforge()
				return
			end
		end
	end
	if next(reforgequeue) ~= nil then
		Reforgerade:RetryReforge()
	else
		self:Print("Reforging Complete")
		waitingToUnforge = {}
		waitingToReforge = {}
		ClearCursor()
		ReforgingFrameItemButton:Click() -- pick up the old item
		ClearCursor() --get rid of it
		currentSlot = 0
	end
	--self:Print("Nothing to reforge")
end

function Reforgerade:RetryReforge()
	local waitTime = max(select(4, GetNetStats())*2, 500) -- Wait for 2x current latentcy or 250ms whichever is larger and do it again
--~ 	self:DebugPrint("Next action in " .. waitTime .. "ms")
	self:ScheduleTimer("ReforgeIt", waitTime/1000)  
end


--removes of the Stormblast from item names so they can be compared to the reforge list
function Reforgerade:FixItemName(itemName)
	itemName = string.gsub(itemName, " of the Stormblast", "")
	itemName = string.gsub(itemName, " of the Windflurry", "")
	itemName = string.gsub(itemName, " of the Windstorm", "")
	itemName = string.gsub(itemName, " of the Zephyr", "")
	return itemName
end

function Reforgerade:RegisterParser(name, func)
	self:DebugPrint("Registering parser " .. name)
	reforgeParsers[name] = func
end

--Stolen from shadowcraft-tools
ReforgeradeInputFrame = {}
ReforgeradeInputFrame.frame = nil
ReforgeradeInputFrame.frame = CreateFrame("Frame", "ReforgeradeInputFrame", UIParent)
ReforgeradeInputFrame.frame:SetBackdrop({
		bgFile = "Interface\\Addons\\Reforgerade\\media\\textures\\blank", --zomg stolen textures too
		edgeFile = "Interface\\Addons\\Reforgerade\\media\\textures\\blank",
		tile = true, tileSize = 16, edgeSize = 1, 
		insets = { left = -1, right = -1, top = -1, bottom = -1 }
})
ReforgeradeInputFrame.frame:SetBackdropColor(0.1,0.1,0.1,1)
ReforgeradeInputFrame.frame:SetBackdropBorderColor(0.6, 0.6, 0.6, 0.5)
ReforgeradeInputFrame.titleString = ReforgeradeInputFrame.frame:CreateFontString("TitleString");
ReforgeradeInputFrame.titleString:SetFont("Interface\\Addons\\Reforgerade\\media\\fonts\\arial.ttf",12)  --eek stolen fonts
ReforgeradeInputFrame.titleString:SetPoint("TOP", ReforgeradeInputFrame.frame, "TOP", -2, -2)
ReforgeradeInputFrame.titleString:SetText("Reforgerade")

ReforgeradeInputFrame.frame:SetWidth(400)
ReforgeradeInputFrame.frame:SetHeight(400)
ReforgeradeInputFrame.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
ReforgeradeInputFrame.frame:Hide()
ReforgeradeInputFrame.frame:SetFrameStrata("DIALOG")

local scrollArea = CreateFrame("ScrollFrame", "ExportScroll", ReforgeradeInputFrame.frame, "UIPanelScrollFrameTemplate")
scrollArea:SetPoint("TOPLEFT", ReforgeradeInputFrame.frame, "TOPLEFT",8, -30)
scrollArea:SetPoint("BOTTOMRIGHT", ReforgeradeInputFrame.frame, "BOTTOMRIGHT", -30, 8)

ReforgeradeInputFrame.editBox = CreateFrame("EditBox", "ExportBox", ReforgeradeInputFrame.frame)
ReforgeradeInputFrame.editBox:SetMultiLine(true)
ReforgeradeInputFrame.editBox:SetMaxLetters(99999)
ReforgeradeInputFrame.editBox:EnableMouse(true)
ReforgeradeInputFrame.editBox:SetAutoFocus(true)
ReforgeradeInputFrame.editBox:SetFontObject(ChatFontNormal)
ReforgeradeInputFrame.editBox:SetWidth(362)
ReforgeradeInputFrame.editBox:SetHeight(400)
ReforgeradeInputFrame.editBox:SetScript("OnEscapePressed", function() ReforgeradeInputFrame.frame:Hide() end)

scrollArea:SetScrollChild(ReforgeradeInputFrame.editBox)

local close = CreateFrame("Button", "CopyCloseButton", ReforgeradeInputFrame.frame, "UIPanelCloseButton")
close:SetPoint("TOPRIGHT", ReforgeradeInputFrame.frame, "TOPRIGHT")

local ReforgeCompareButton = CreateFrame("Button", "ReforgeCompareButton", ReforgeradeInputFrame.frame, "OptionsButtonTemplate")
ReforgeCompareButton:SetPoint("BOTTOMRIGHT", ReforgeradeInputFrame.frame, "BOTTOM")
ReforgeCompareButton:SetText("Compare")
ReforgeCompareButton:SetScript("OnClick", function() Reforgerade:ReforgeScan(false) end)

local ReforgeItButton = CreateFrame("Button", "ReforgeItButton", ReforgeradeInputFrame.frame, "OptionsButtonTemplate")
ReforgeItButton:SetPoint("BOTTOMLEFT", ReforgeradeInputFrame.frame, "BOTTOM")
ReforgeItButton:SetText("Reforge It")
ReforgeItButton:SetScript("OnClick", function() Reforgerade:ReforgeIt() end)